\begin{exercise}
\ifsolution\else
The file <<simpinh.f>> (listed in
Figure~\ref{fig:simpinh} and available as part of the <<fomega>>
distribution) contains a simple implementation
of inheritance without the pseudovariable <<self>>.  
\begin{figure}
\verbatimfile{simpinh.f}
\caption{The file {\tt simpinh.f}\hspace{2in}}\label{fig:simpinh}
\end{figure}
\fi
Copy the file <<simpinh.f>> and add a simple <<slide>> method to the
definitions of points; this <<slide>> should be inherited by the
definition of colored points.  For the moment, you will not be able to
get exactly the same 
behavior as in the Smalltalk implementation of slideable points, where
the <<slide>> method defined by the point class calls the overriden
<<setX>> method of colored points.  Define <<slide>> here so that it
sets the instance variable <<x>> directly.
\end{exercise}

\ifsolution
\noindent Should be easy.
\fi

\breakifnearbottom
\begin{exercise}
\ifsolution\else
Adding <<self>> is actually quite easy.  Each class is extended with
an extra parameter for the ``<<self>> methods.  For example, the
<<pointClass>> from the lecture becomes:
%%% !besilent;
> Point = ObjectType(Rep) with
>             getX: Int,
>             setX: Int->Rep
>           end;
> PointR = {x:Int};
> CPointR = {x:Int, color:Color};
%%% !beloud;
> pointClass = 
>    fun(FinalR)
>    fun(get: FinalR->PointR)
>    fun(put: FinalR->PointR->FinalR)
>    fun(self: PointM(FinalR))
>      {getX = fun(s:FinalR) (get s).x,
>       setX = fun(s:FinalR) fun(i:Int) put s {x=i}
>      } :<PointM FinalR>;
Then building a new object from a class involves using a fixed point
to build the actual method vector:
> p = < PointR,
>       {state = {x=5},
>        methods = rec (PointM(PointR))
>                    fun(self:PointM(PointR))
>                      pointClass
>                        PointR
>                        (fun(s:PointR) s)
>                        (fun(s:PointR) fun(s':PointR) s')
>                        self}
>     >: Object(PointM);
\fi
Extend this example with a <<slide>> method that uses the <<setX>> and
<<getX>> methods from <<self>>.
\end{exercise}

\ifsolution
> SlidePointM = Fun(R:*) {|setX:R->Int->R, getX:R->Int, slide:R->Int->R|};
> slidePointClass = 
>    fun(FinalR)
>    fun(get: FinalR->PointR)
>    fun(put: FinalR->PointR->FinalR)
>    fun(self: PointM(FinalR))
>      {getX = fun(s:FinalR) (get s).x,
>       setX = fun(s:FinalR) fun(i:Int) put s {x=i},
>       slide = fun(s:FinalR) fun(i:Int) 
>                 self.setX s (plus (self.getX s) i)
>      } :<SlidePointM FinalR>;
\fi

\begin{exercise}
By analogy with the previous exercise, 
show how to add <<self>> to the definitions of <<Class>>, <<new>>, and
<<extend>> in the file <<simpinh.f>>.
\end{exercise}

\ifsolution
> Class = 
>   Fun(M:*->*)
>   Fun(R:*)
>     All(FinalR)
>          (FinalR->R)
>       -> (FinalR->R->FinalR)
>       -> M(FinalR)
>       -> M(FinalR);
> 
> new =
>   fun(M<Top(*->*))
>   fun(R)
>   fun(s:R)
>   fun(cl: Class M R)
>     < R, 
>       {state = s,
>        methods = rec (M(R))
>                    fun(self:M(R)) 
>                       cl
>                         R
>                         (fun(s:R) s)
>                         (fun(s:R) fun(s':R) s')
>                         self}
>     >: Object(M);
> 
> extend = 
>   fun(SuperM < Top(*->*))                     % superclass interface
>   fun(SuperR)                                 % superclass representation
>   fun(NewM < SuperM)                          % new class interface
>   fun(NewR)                                   % new class representation
>   fun(superClass: Class SuperM SuperR)        % the superclass        
>   fun(build:                                  % "increment" fun, with args...
>           All(FinalR)                            % final representation
>           (FinalR->NewR) ->                      % extractor
>           (FinalR->NewR->FinalR) ->              % overwriter
>           (SuperM(FinalR)) ->                    % superclass methods
>           (NewM(FinalR)) ->                      % self methods
>           (NewM(FinalR)))                        % ...returning new methods
>   fun(get: NewR->SuperR)                      % new->super extractor
>   fun(put: NewR->SuperR->NewR)                % new<-super overwriter
>    % Build the extended class...
>    (fun(FinalR)                          
>     fun(g: FinalR->NewR)
>     fun(p: FinalR->NewR->FinalR)
>     fun(self: NewM(FinalR))
>       build FinalR g p
>          (superClass FinalR
>            (fun(s:FinalR) get(g(s))) 
>            (fun(s:FinalR) fun(s':SuperR) p s (put (g s) s'))
>            self)
>          self);
\fi

\begin{exercise}
Use <<extend>> to give a slideable version of the colored point class
that inherits the slide method from slidePointClass.  Show that
instances of this class have the same behavior as their Smalltalk
counterparts (i.e.~the <<slide>> method inherited from points uses the
overriden <<setX>> method of colored points). 
\end{exercise}

\ifsolution
> SlideCPointM = Fun(R:*) {|setX:R->Int->R, getX:R->Int,
>                           setC:R->Color->R, getC:R->Color,
>                           slide:R->Int->R|};
> CPointR = {|x:Int, color:Color|};

> slideCPointClass =
>   extend SlidePointM PointR SlideCPointM CPointR slidePointClass
>   (fun(FinalR)
>    fun(get: FinalR->CPointR)
>    fun(put: FinalR->CPointR->FinalR)
>    fun(super: SlidePointM(FinalR))
>    fun(self: SlideCPointM(FinalR))
>     {setX = fun(s:FinalR) fun(i:Int) 
>               put s {x=i,color=blue},
>      getX = super.getX,
>      slide = super.slide,
>      setC = fun(s:FinalR) fun(c:Color) 
>                put s {x=(get s).x, color=c},
>      getC = fun(s:FinalR) (get s).color})
>   (fun(s:CPointR) {x=s.x})
>   (fun(s:CPointR) fun(s':PointR) {x=s'.x, color=s.color});
After loading the functions for sending messages to points), we
can test this implementation as before:
%%%!besilent;
> SlidePoint'setX =
>  fun(M<SlidePointM)
>    fun(p:Object(M))
>      fun(i:Int)
>        open p as <R,r> in
>          <R, 
>           {state = r.methods.setX r.state i,
>            methods = r.methods}
>          >: Object(M)
>       end;
> SlidePoint'slide =
>  fun(M<SlidePointM)
>    fun(p:Object(M))
>      fun(i:Int)
>        open p as <R,r> in
>          <R, 
>           {state = r.methods.slide r.state i,
>            methods = r.methods}
>          >: Object(M)
>       end;
> SlideCPoint'setC =
>  fun(M<SlideCPointM)
>    fun(p:Object(M))
>      fun(c:Color)
>        open p as <R,r> in
>          <R,
>           {state = r.methods.setC r.state c,
>            methods = r.methods}
>          >: Object(M)
>       end;
> SlidePoint'getX =
>  fun(M<SlidePointM)
>    fun(p:Object(M))
>       open p as <R,r> in
>         r.methods.getX r.state
>       end;
> SlideCPoint'getC =
>  fun(M<SlideCPointM)
>    fun(p:Object(M))
>       open p as <R,r> in
>         r.methods.getC r.state
>       end;
%%%!beloud;
> c = new SlideCPointM CPointR {x=5,color=red} slideCPointClass;
> c1 = SlidePoint'setX SlideCPointM c 3;
> c2 = SlideCPoint'setC SlideCPointM c1 yellow;
> c3 = SlidePoint'slide SlideCPointM c2 3;
> SlideCPoint'getC SlideCPointM c3;
\fi

\ifambitious
\begin{exercise} (Optional)
\finish{Inheritance with private instance variables.}
\end{exercise}
\fi

