functor Synth(
	    structure Globals: GLOBALS
	    structure Typ: TYP
	    structure TEnv: TENV
	    structure TypHelper: TYPHELPER
	    structure Trm: TRM
	    structure TrmHelper: TRMHELPER
	    structure Leq: LEQ
	    structure Simplify: SIMPLIFY
	    sharing Typ.Globals = Globals
	        and Trm.Typ = Typ
	        and Leq.Typ = Typ
	        and TypHelper.Typ = Typ
	        and TrmHelper.Trm = Trm
	        and Trm.TEnv = TEnv
	        and Leq.TEnv = TEnv
	        and Typ.Id = Globals.Id
	        and Simplify.TEnv = TEnv
	        and Simplify.Typ = Typ
	    ) : SYNTH = struct

structure Globals = Globals
structure Typ = Typ
structure Trm = Trm
structure Leq = Leq
structure TEnv = TEnv
open Globals
open TEnv
open Typ
open Trm

val querytenvdepth = ref(0)
fun setquerytenv te = (querytenvdepth := (domsize te))

val DEBUG = createflag ["synth", "Synth"] false

val synthie = createflag ["synthie"] false

fun warning s = write ("Warning: " ^ s ^ "\n");

fun arrowbasis te t = let 
  fun ab t =
    case explode(t) of
          EXPTVAR v => 
   	    (lookup_and_relocate te v 
   	        (fn b => ab b)
   	        (fn _ => raise TEnv.WrongKindOfId(te,v,"tvar or abbrev"))
	        (fn b => ab b))
        | EXPARROW a => 
            [a]
        | EXPALL _ => 
            []
        | EXPSOME _ => 
            []
        | EXPMEET ts =>
            mapappend ab ts
        | EXPOBJ _ => 
            []
        | EXPMINUS (t,_) => 
            ab t
  in ab t
  end

fun allbasis te t = let 
  fun ab t =
    case explode(t ) of
          EXPTVAR v => 
   	    (lookup_and_relocate te v 
		(fn b => ab b)
	        (fn _ => raise TEnv.WrongKindOfId(te,v,"tvar or abbrev"))
	        (fn b => ab b))
        | EXPARROW _ => 
            []
        | EXPALL a => 
            [a]
        | EXPSOME _ => 
            []
        | EXPMEET ts =>
            mapappend ab ts
        | EXPOBJ _ => 
            []
        | EXPMINUS (t,_) => 
            ab t
  in ab t
  end

fun somebasis te t = let 
  fun ab t =
    case explode(t ) of
          EXPTVAR v => 
   	    (lookup_and_relocate te v 
		(fn b => ab b)
	        (fn _ => raise TEnv.WrongKindOfId(te,v,"tvar or abbrev"))
	        (fn b => ab b))
        | EXPARROW _ => 
            []
        | EXPALL _ => 
            []
        | EXPSOME a => 
            [a]
        | EXPMEET ts =>
            mapappend ab ts
        | EXPOBJ _ => 
            []
        | EXPMINUS (t,_) => 
            ab t
  in ab t
  end

fun objbasis te t = let 
  fun ab t =
    case explode(t ) of
          EXPTVAR v => 
   	    (lookup_and_relocate te v 
		(fn b => ab b)
	        (fn _ => raise TEnv.WrongKindOfId(te,v,"tvar or abbrev"))
	        (fn b => ab b))
        | EXPARROW _ => 
            []
        | EXPALL _ => 
            []
        | EXPSOME _ => 
            []
        | EXPMEET ts =>
            mapappend ab ts
        | EXPOBJ a => 
            [a]
        | EXPMINUS (t1,ids) => 
            filter (fn (i,_) => not (memq Id.== ids i)) (objbasis te t1)
  in ab t
  end

fun synth' te (VAR(_,v)) = TEnv.lookup_and_relocate_binding te v
  | synth' te (ABS({name=i},t,body)) = 
      let val t_body = synth (push_binding te i t) body
          val t' = relocate ~1 t_body
      in arrow t t'
      end
  | synth' te (APP(_,trm1,trm2)) = 
      let val t1 = synth te trm1
          and t2 = synth te trm2
          val basis = arrowbasis te t1
          fun collect_apps [] =
                []
            | collect_apps ((tb1,tb2)::tl) = 
                if Leq.leq te t2 tb1
                   then tb2::(collect_apps tl)
                   else (collect_apps tl)
          val ts = collect_apps basis
      in meet ts 
      end
  | synth' te (TABS({name=i},t,body)) = 
      let val t' = synth (push_bound te i t) body
      in all i t t'
      end
  | synth' te (TAPP(_,body,t)) =
      let val t_body = synth te body
          val basis = allbasis te t_body
          fun collect_apps [] =
                []
            | collect_apps ((_,t1,t2)::tl) = 
                if Leq.leq te t t1
                   then (tsubst_top t t2)::(collect_apps tl)
                   else (collect_apps tl)
          val ts = collect_apps basis
      in meet ts
      end
  | synth' te (FOR({name=i},ts,body)) =
      let fun f t = 
              let val tb = synth (push_abbrev te i t) body
                  val tb' = tsubst_top t tb
              in tb'
              end
      in meet (map f ts)
      end
  | synth' te (WITH(body,l,v)) = let
      val t_body = synth te body
      val t_v = synth te v
      in meet [minus t_body [l], obj l t_v]
      end
  | synth' te (PROJ(body,l)) = let
      val t_body = synth te body
	  (* $$$ would be more efficient to make objbasis take l as parameter *)
      val basis = objbasis te t_body
      fun collect_apps [] =
                []
        | collect_apps ((i,t2)::tl) = 
                if Id.== i l
                   then t2::(collect_apps tl)
                   else (collect_apps tl)
      val ts = collect_apps basis
      in meet ts
      end
  | synth' te (OPEN(a,x,p,body)) = let
      val t_p = synth te p
      val basis = somebasis te t_p
      fun collect_apps [] =
                []
        | collect_apps ((i,t1,t2)::tl) = let
		val rest = collect_apps tl
                val t' = synth (push_binding (push_bound te a t1) x t2) body
                in 
                   if occurs_free_in 1 t'
                     then (warning "Representation type would escape";
                     	   rest)
                     else (relocate ~2 t')::rest
                end
      val ts = collect_apps basis
      in meet ts
      end
  | synth' te (PACK(e,t,hidden)) = let
      (* Simplified and very incomplete treatment: *)
      val s = synth te e
      in case explode(t ) of
	       EXPTVAR v => 
		 (lookup_and_relocate te v 
		     (fn b => NS)
		     (fn _ => raise TEnv.WrongKindOfId(te,v,"tvar or abbrev"))
		     (fn t' => synth' te (PACK(e,t',hidden))))  (* ugly? *)
	     | EXPARROW _ => 
		 NS
	     | EXPALL _ => 
		 NS
	     | EXPSOME (i,t1,t2) => 
		 if (Leq.leq te hidden t1)
		    andalso (Leq.leq te s (Typ.tsubst_top hidden t2))
		    then t
		    else NS
	     | EXPMEET _ => 
		 (warning "Treatment of PACK is incomplete!";
		  NS)
	     | EXPOBJ _ => 
		 NS
	     | EXPMINUS _ => 
		 (warning "Treatment of PACK is incomplete!";
		  NS)
      end
  | synth' te (NOTHING) = 
      NS
  | synth' te (QUERY(tag,body)) = let
      val res = synth te body
      val noabbs = (domsize te) - (!querytenvdepth)
      fun wabs 0 = true
        | wabs n = let 
            val first = wabs (n-1)
            val curr = (noabbs - n)
            in lookup_and_relocate te curr
                 (fn _ => first)
                 (fn _ => first)
                 (fn t => 
                    (if (not first) 
		       then (write ", ";
			     break false 0)
		       else ();
		     write (Id.tostr (TEnv.lookup_name te curr));
		     write "=";
		     TypHelper.prt te t;
		     false))
            end
      in write "  ";
         setb ();
         if (wabs noabbs)
              then ()
              else (break true ~2;
                    write "  => ");
         write tag;
         write ": ";
         if (!synthie) 
           then (TypHelper.prt te res;
           	 ubreak ~5;
           	 write "i.e. ")
           else ();
         TypHelper.prt te (Simplify.simplify te res);
         endb ();
         write "\n";
         res
      end

and synth te e =
  wrap (DEBUG) "synth"
       (fn () => synth' te e)
       (fn () => 
       	  (setb ();
       	   TrmHelper.prt te e;
       	   break true ~5;
       	   write "  in ";
       	   TypHelper.prt_tenv te;
       	   endb ()))
       (fn t => 
          (setb ();
           TypHelper.prt te t;
           break true ~2;
           write "  i.e. ";
           Typ.prt_exactly t;
           endb ()))

end
