% %%%%%%%%%%%%%%%%%%%%%
% %%%    Objects    %%%
% %%%%%%%%%%%%%%%%%%%%%

% Given a "method vector" type operator M, build the type of objects with 
% methods M:
Object = Fun(M:*->*) Some(A<Top1) {|state:A, methods:M(A)|};


% %%%%%%%%%%%%%%%%%%%%%%%%
% %%%    Extractors    %%%
% %%%%%%%%%%%%%%%%%%%%%%%%

% An "extractor" from S to T provides a way of extracting an element of
% T from an element of S and a way of installing a new value for the "T
% part of an S" into an element of S.  (This is the functional
% analog of a pointer into a state vector.)
Extractor = Fun(S:*) Fun(T:*) {|get: S->T, put: S->T->S|};

idExtractor = 
  fun(SelfR<Top1) 
    {get = fun(s:SelfR) s, 
     put = fun(s:SelfR) fun(t:SelfR) t};

composeExtractors = 
  fun(T1<Top1) fun(T2<Top1) fun(T3<Top1)
    fun(e1: Extractor T1 T2)
      fun(e2: Extractor T2 T3)
        {get = fun(t1:T1) e2.get (e1.get t1),
         put = fun(t1:T1) fun(t3:T3) 
                 e1.put t1 (e2.put (e1.get t1) t3)};

Pair = Fun(T1:*) Fun(T2:*) {|fst:T1, snd:T2|};

pair = 
  fun(T1<Top1) 
  fun(T2<Top1) 
  fun(t1:T1)
  fun(t2:T2)
    {fst=t1, snd=t2};

fstExtractor = 
 fun(T1<Top1) fun(T2<Top1) 
   {get = fun(p: Pair T1 T2) p.fst, 
    put = fun(p: Pair T1 T2) fun(t:T1) {fst=t, snd=p.snd} };

sndExtractor = 
 fun(T1<Top1) fun(T2<Top1) 
   {get = fun(p: Pair T1 T2) p.snd, 
    put = fun(p: Pair T1 T2) fun(t:T2) {fst=p.fst, snd=t} };


% %%%%%%%%%%%%%%%%%%%%%%
% %%%     Classes    %%%
% %%%%%%%%%%%%%%%%%%%%%%

ClassMethods =
  Fun(SelfM:*->*)
  Fun(SelfR:*)
    All(FinalR<Top1)
    (Extractor FinalR SelfR) ->
    (SelfM(FinalR)) ->
    (SelfM(FinalR));

Class =
  Fun(SelfM:*->*) 
    Some(SelfR<Top1)
      {|localstate: SelfR,
        buildM: ClassMethods SelfM SelfR|};
     
new =
  fun(SelfM < Top2) 
  fun(selfClass: Class SelfM)
    open selfClass as <SelfR,selfData> in
      <SelfR,
       {state = selfData.localstate,
        methods = 
          rec (SelfM(SelfR))
            (fun(self:SelfM(SelfR))
              selfData.buildM SelfR (idExtractor SelfR) self)} 
       > : Object(SelfM)
    end;

extend = 
  fun(SuperM < Top2)
  fun(SelfM < SuperM)
  fun(SelfDeltaR < Top1)
  fun(superClass: Class SuperM)
  fun(deltastate: SelfDeltaR)
  fun(build: All(FinalR<Top(*))
               (FinalR->SelfDeltaR) ->
               (FinalR->SelfDeltaR->FinalR) ->
               (SuperM(FinalR)) -> 
               (SelfM(FinalR)) -> 
               (SelfM(FinalR)))
    open superClass as <SuperR,superData> in 
      <Pair SuperR SelfDeltaR,
       {localstate = pair SuperR SelfDeltaR
                          (superData.localstate) deltastate,
        buildM = 
         fun(FinalR<Top(*))
         fun(e: Extractor FinalR (Pair SuperR SelfDeltaR))
         fun(self: SelfM(FinalR))
           let esnd = composeExtractors 
                         FinalR (Pair SuperR SelfDeltaR) SelfDeltaR
                         e (sndExtractor SuperR SelfDeltaR)
           in build FinalR esnd.get esnd.put
             (superData.buildM FinalR 
                (composeExtractors FinalR (Pair SuperR SelfDeltaR) SuperR e
                                  (fstExtractor SuperR SelfDeltaR))
                self)
             self
           end}
       > : Class SelfM
     end;

Thing = 
  Objecttype(Rep) with end;

thingClass =
  <Unit,
   {localstate = unit,
    buildM = fun(FinalR<Top1)
             fun(e: Extractor FinalR Unit)
             fun(self: ThingM(FinalR))
               {} }
  > : Class ThingM;

printer 
  fun(t:Object(Fun(X)X->Top(*))) write "<object>";
