functor TopLevel(
	 structure Globals: GLOBALS
	 structure TEnv: TENV
	 structure TypHelper: TYPHELPER
	 structure TrmHelper: TRMHELPER
	 structure ParseRes: PARSERES
	 structure Synth: SYNTH
	 structure Typ: TYP
	 structure Eval: EVAL
	 structure Object: OBJECT
	 sharing TypHelper.Globals = Globals
	     and ParseRes.TypHelper = TypHelper
	     and ParseRes.TrmHelper = TrmHelper
	     and Synth.TEnv = TEnv
	     and TypHelper.TEnv = TEnv
	     and TEnv.Typ = Typ
	     and Synth.TypHelper = TypHelper
	     and Synth.Trm = TrmHelper.Trm
	     and Eval.Trm = TrmHelper.Trm
	     and Eval.Id = Globals.Id
	     and Eval.Globals = Globals
	     and Eval.TEnv = TEnv
	     and Object.Eval = Eval
	     and Object.TypHelper = TypHelper
	     and Object.TrmHelper = TrmHelper
	     and Object.Id = Globals.Id
	) : TOPLEVEL = struct

structure ParseRes = ParseRes

open Globals;
open ParseRes;

val recap = createflag ["recap"] false
val origterms = createflag ["origterms"] false

(* NB: same name as in object.sml *)
val eval = createflag ["eval"] true

type TopLevelData = TEnv.T * Eval.env * Eval.printercontext

fun init () = (TEnv.empty, Eval.empty_env, Eval.empty_printercontext)

fun do_command (tld as (te,env,pctx)) 
               (continue: TopLevelData->'a) 
               (error: TopLevelData->'a) 
               cmd =  
    (case cmd of
         Type_Assumption(i,pt) => let
       	   val t = TypHelper.debruijnify te pt
           in if !recap 
                then (write (Id.tostr i); write " < ";
                      TypHelper.prt_outer te t; write "\n")
	        else ();
       	      continue (TEnv.push_bound te i t, 
       	                Eval.push_env_null env i,
       	                Eval.deepen_printers pctx)
           end
       | Type_Abbrev(i,pt) => let 
           val t = TypHelper.debruijnify te pt
           val k = Synth.ksynth te t
           in write (Id.tostr i); write " "; 
              setb();
              if !recap 
                then (write "= "; TypHelper.prt_outer te t; write "\n")
	        else ();
	      write ": ";
	      TypHelper.kprt te k;
	      endb();
	      write "\n";
	      continue (TEnv.push_abbrev te i t, 
	                Eval.push_env_const env i,
	                Eval.deepen_printers pctx)
           end
       | Term_Def(i,ptrm) => let 
           val trm = TrmHelper.debruijnify te ptrm
       	   val _ = Synth.setquerytenv te
       	   val otyp = Synth.synth te trm
       	   val typ = Synth.squash te otyp
      	   val norm = if !eval
      	                 then Eval.eval te env i trm
      	                 else Eval.dummycode
      	   val noprint = (Id.== i (Id.intern "it"))
           in (* It would be cleaner to make the grammar return an 
                 string Option instead of casing on the name *) 
              if noprint then () 
                else write (Id.tostr i);
              setb();
              if not (!eval) then () else
                 (if noprint then () 
                    else write " = ";
                  setb(); Eval.eval_print te pctx typ i; endb());
	      ubreak 0;
              write " : ";
              TypHelper.prt_outer te typ;
              endb();
              write "\n";
              continue (TEnv.push_abbrev 
              		  (TEnv.push_binding te i typ)
              		  (Id.intern "It")
              		  (Typ.relocate 1 typ),
              		Eval.push_env_null 
              		  (Eval.push_env_binding env i norm)
              		  (Id.intern "It"),
              		Eval.deepen_printers_by pctx 2)
                      end
       | Term_Assumption(i,ptp) => let 
           val typ = TypHelper.debruijnify te ptp
           in if !recap
                then (write (Id.tostr i); write " : "; 
                      TypHelper.prt_outer te typ; write "\n")
		else ();
              continue (TEnv.push_abbrev 
              		  (TEnv.push_binding te i typ)
              		  (Id.intern "It")
              		  (Typ.relocate 1 typ),
              		Eval.push_env_null 
              		  (Eval.push_env_binding env i Eval.dummycode)
              		  (Id.intern "It"),
              		Eval.deepen_printers_by pctx 2)
           end
       | ObjectTypeDecl(name,selfid,method_decls) => let
           val (new_te, new_env, howmany) = 
                 Object.build_objecttype name selfid method_decls te env
           in continue (new_te, new_env, Eval.deepen_printers_by pctx howmany)
           end
       | ClassDecl(name,selftype,supername,supertype,
                   instancevars,
                   inherits,virtual,method_bodies) => let
           val (new_te, new_env, howmany) = 
                 Object.build_classdecl 
                     name selftype supername supertype 
                     instancevars inherits virtual method_bodies te env
           in continue (new_te, new_env, Eval.deepen_printers_by pctx howmany)
           end
       | PrinterDecl(ptrm) => let
           val trm = TrmHelper.debruijnify te ptrm
       	   val _ = Synth.setquerytenv te
       	   val otyp = Synth.synth te trm
       	   val typ = Synth.squash te otyp
       	   val name = Id.new_from "printer"
      	   val norm = Eval.eval te env name trm
           in continue (te, env, Eval.define_printer pctx typ norm)
           end
       | Leq(pt1,pt2) => let 
           val t1 = TypHelper.debruijnify te pt1
           val t2 = TypHelper.debruijnify te pt2
           in 
              if Synth.leq te t1 t2
                then write "Yes\n"
                else write "No\n";
	      continue tld
           end
       | Eqv(pt1,pt2) => let 
           val t1 = TypHelper.debruijnify te pt1
           val t2 = TypHelper.debruijnify te pt2
           in 
              if (Synth.eqv te t1 t2) 
                then write "Yes\n"
                else write "No\n";
	      continue tld
           end
       )
    handle 
      TEnv.WrongKindOfId(te,i,which) => 
      	  (write ("Wrong kind of identifier: "^ (makestring i) ^" ("
      	  	  ^ which ^ " expected)\nin "); 
      	   TypHelper.prt_tenv te;
       	   error tld)
    | TEnv.UnknownId(s) =>
          (write ("Unbound id: " ^ s ^ "\n");
           error tld)
    | Object.BadObjDecl(s) =>
          (write ("Bad object type: " ^ s ^ "\n");
           error tld)
    | Synth.IllTyped(f) =>
          (write "Typechecking failure: \n";
           write "  "; setb(); f(); endb(); write "\n";
           error tld)
    | Synth.SynthFailure(f) =>
          (write "Subtyping or kinding failure: \n";
           write "  "; setb(); f(); endb(); write "\n";
           error tld)
    | Eval.Eval(f) =>
          (write "Evaluation failure: \n";
           write "  "; setb(); f(); endb(); write "\n";
           error tld)

end


